/****************************************************************
 *                                                              *
 * File          : nvector.c                                    *
 * Programmers   : Scott D. Cohen and Alan C. Hindmarsh @ LLNL  *
 * Last Modified : 8 May 1998                                   *
 *--------------------------------------------------------------*
 *                                                              *
 * This is the implementation file for a generic VECTOR         *
 * package. It contains the implementation of the N_Vector      *
 * kernels listed in nvector.h.                                 *
 * Modified for MPI.                                            *
 *                                                              *
 ****************************************************************/

/*......................................................................


                            LEGAL NOTICES


1. NOTICE

This work was performed at the University of California, Lawrence
Livermore National Laboratory (UC LLNL) under contract no.
W-7405-ENG-48 (Contract 48) between the U.S. Department of Energy
(DOE) and The Regents of the University of California (the University)
for the operation of UC LLNL.  Copyright is reserved to the University
for purposes of controlled dissemination, commercialization through
formal licensing, or other disposition under terms of Contract 48; DOE
policies, regulations, and orders; and U.S. statutes.  The rights of
the Federal Government are reserved under Contract 48 subject to the
restrictions agreed upon by the DOE and University as allowed under
DOE Acquisition Letter 97-1.


2. DISCLAIMER

This work was prepared as an account of work sponsored by an agency of
the United States Government.  Neither the United States Government
nor the University of California nor any of their empolyees makes any
warranty, express or implied, or assumes any liability or
responsibility for the accuracy, completeness, or usefulness of any
information, apparatus, product, or process disclosed, or represents
that its use would not infringe privately owned rights.  Reference
herein to any specific commercial products, process, or service by
trade name, trademark, manufacturer, or otherwise, does not
necessarily constitute or imply its endorsement, recommendation, or
favoring by the United States Government or the University of
California.  The views and opinions of authors expressed herein do not
necessarily state or reflect those of the United States Government or
the University of California, and shall not be used for advertising or
product endorsement purposes.


3. LICENSING REQUIREMENTS

Permission is hereby granted to use and copy this software and
documentation for internal non-commercial purposes only, provided that
all UC LLNL identification in the user interface remains unchanged.
Any use, reproduction, modification, or distribution of this softwre
or documentation for commercial purposes requires a license from the
University.  Contact: Lawrence Livermore National Laboratory,
Industrial Partnerships and Commercialization Office, P.O. Box 808,
L-795, Livermore, CA 94551.

......................................................................*/



#include <stdio.h>
#include <stdlib.h>
#include "nvector.h"
#include "llnltyps.h"
#include "llnlmath.h" 


#define ZERO RCONST(0.0)
#define ONE  RCONST(1.0)

/* Error message */

#define BAD_N1  "PVecInitMPI-- Sum of local vector lengths differs from "
#define BAD_N2  "input global length. \n\n"
#define BAD_N    BAD_N1 BAD_N2

/* Private Helper Prototypes */

static real PVecAllReduce(real d, int op, machEnvType machEnv);
/* Reduction operations add/max/min over the processor group */
static void VCopy(N_Vector x, N_Vector z); /* z=x */
static void VSum(N_Vector x, N_Vector y, N_Vector z); /* z=x+y */
static void VDiff(N_Vector x, N_Vector y, N_Vector z); /* z=x-y */
static void VNeg(N_Vector x, N_Vector z); /* z=-x */
/* z=c(x+y) */
static void VScaleSum(real c, N_Vector x, N_Vector y, N_Vector z);
/* z=c(x-y) */
static void VScaleDiff(real c, N_Vector x, N_Vector y, N_Vector z); 
static void VLin1(real a, N_Vector x, N_Vector y, N_Vector z); /* z=ax+y */
static void VLin2(real a, N_Vector x, N_Vector y, N_Vector z); /* z=ax-y */
static void Vaxpy(real a, N_Vector x, N_Vector y); /* y <- ax+y */
static void VScaleBy(real a, N_Vector x); /* x <- ax */


/********************* Exported Functions ************************/


void *PVecInitMPI(MPI_Comm comm,  integer local_vec_length, 
		  integer global_vec_length, int *argc, char ***argv)
{
  int initflag, initerr;
  integer n, Nsum;
  machEnvType env;

  /* Create structure env and begin loading it */
  env = (machEnvType) malloc (sizeof *env);
  if (env == NULL) return(NULL);

  env->local_vec_length = local_vec_length;
  env->global_vec_length = global_vec_length;

  MPI_Initialized(&initflag);
  if (!initflag) {
    initerr = MPI_Init(argc,argv);
    if (initerr != MPI_SUCCESS) return(NULL);
    }
  env->init_by_user = initflag;

  env->comm = (comm == NULL ? MPI_COMM_WORLD : comm); 

  /* If this PE is inactive, return now */
  if (local_vec_length <= 0) return(env);

  /* Compute global length as sum of local lengths */
  n = local_vec_length;
  MPI_Allreduce(&n, &Nsum, 1, PVEC_INTEGER_MPI_TYPE, MPI_SUM, comm);
  env->global_vec_length = Nsum;

  /* Check input global length against computed value */
  if (Nsum != global_vec_length) {
    printf(BAD_N);
    PVecFreeMPI(env);
    return(NULL);
    } 

  /* Return the pointer env */
  return(env);
}


void PVecFreeMPI(void *machEnv)
{
  machEnvType env;

  env = (machEnvType) machEnv;
  if (env == NULL) return;

  if (!(env->init_by_user)) MPI_Finalize();

  free(machEnv);
}

 
N_Vector N_VNew(integer N, machEnvType machEnv)
{
  N_Vector v; 
  int N_local, N_global; 
  int i, j, ns;
  SensData sdata; 
  N_Vector *ptr; 

  if (N <= 0) return(NULL);
  if (machEnv == NULL) return(NULL);

  N_local = machEnv->local_vec_length;
  N_global = machEnv->global_vec_length;

  if (N == N_global) {
    v = (N_Vector) malloc(sizeof *v);
    if (v == NULL) return(NULL);
  
    v->data = (real *) malloc(N_local * sizeof(real));
    if (v->data == NULL) {
      free(v);
      return(NULL);
    }

    v->length = N_local;
    v->global_length = N_global;
    v->machEnv = machEnv;

    ptr = (N_Vector *) malloc(sizeof(N_Vector));
    if (ptr == NULL) {
      free(v->data);
      free(v);
      return(NULL);
    }
    *ptr = v;
    v->subptr = ptr;
    return(v);    
  }

  /*** N larger than N_global ***/

  sdata = (SensData) machEnv->s_data;  
  ns = sdata->ns;                  

  v = (N_Vector) malloc(sizeof *v); 
  if (v == NULL) return(NULL); 

  v->data = (real *) malloc((ns+1)*N_local*sizeof(real)); 
  if (v->data == NULL) { 
    free(v); 
    return(NULL); 
  } 

  v->length = N_local*(ns+1); 
  v->global_length = N_global*(ns+1); 
  v->machEnv = machEnv; 

  ptr = (N_Vector *) malloc((ns+1)*sizeof(N_Vector));
  if (ptr == NULL) {
    free(v->data);
    free(v);
    return(NULL);
  }
  for (i=0; i <= ns; i++){ 
    ptr[i] = (N_Vector) malloc(sizeof(**ptr)); 
    if (ptr[i] == NULL){ 
      for (j=(i-1); j >= 0; j--) {
	free(ptr[j]);
      }
      free(ptr);
      free(v->data);
      free(v); 
      return(NULL); 
    }
    (ptr[i])->data = v->data + (i*N_local); 
    (ptr[i])->length = N_local; 
    (ptr[i])->global_length = N_global; 
    (ptr[i])->machEnv = machEnv; 
    (ptr[i])->subptr = NULL; 
  } 
  v->subptr = ptr;

  return(v);
}


void N_VFree(N_Vector x)
{
  int N, N_global;
  int i, ns;
  machEnvType machEnv;
  SensData sdata;
  N_Vector *ptr;

  N = x->global_length;

  machEnv = (machEnvType) x->machEnv;
  N_global = machEnv->global_vec_length;

  if (N == N_global) {
    free(x->data);
    free(x->subptr);
    free(x);
    return;
  }

  /*** N larger than N_global ***/

  sdata = (SensData) machEnv->s_data;
  ns = sdata->ns;

  ptr = (N_Vector *) x->subptr;
  for (i=0; i <= ns; i++) {
    free(ptr[i]);
  }
  free(x->subptr); 
  free(x->data);         
  free(x);
  return;
}


void N_VLinearSum(real a, N_Vector x, real b, N_Vector y, N_Vector z)
{
  integer i, N;
  real c, *xd, *yd, *zd;
  N_Vector v1, v2;
  boole test;

  if ((b == ONE) && (z == y)) {    /* BLAS usage: axpy y <- ax+y */
    Vaxpy(a,x,y);
    return;
  }

  if ((a == ONE) && (z == x)) {    /* BLAS usage: axpy x <- by+x */
    Vaxpy(b,y,x);
    return;
  }

  /* Case: a == b == 1.0 */

  if ((a == ONE) && (b == ONE)) {
    VSum(x, y, z);
    return;
  }

  /* Cases: (1) a == 1.0, b = -1.0, (2) a == -1.0, b == 1.0 */

  if ((test = ((a == ONE) && (b == -ONE))) || ((a == -ONE) && (b == ONE))) {
    v1 = test ? y : x;
    v2 = test ? x : y;
    VDiff(v2, v1, z);
    return;
  }

  /* Cases: (1) a == 1.0, b == other or 0.0, (2) a == other or 0.0, b == 1.0 */
  /* if a or b is 0.0, then user should have called N_VScale */

  if ((test = (a == ONE)) || (b == ONE)) {
    c = test ? b : a;
    v1 = test ? y : x;
    v2 = test ? x : y;
    VLin1(c, v1, v2, z);
    return;
  }

  /* Cases: (1) a == -1.0, b != 1.0, (2) a != 1.0, b == -1.0 */

  if ((test = (a == -ONE)) || (b == -ONE)) {
    c = test ? b : a;
    v1 = test ? y : x;
    v2 = test ? x : y;
    VLin2(c, v1, v2, z);
    return;
  }

  /* Case: a == b */
  /* catches case both a and b are 0.0 - user should have called N_VConst */

  if (a == b) {
    VScaleSum(a, x, y, z);
    return;
  }

  /* Case: a == -b */

  if (a == -b) {
    VScaleDiff(a, x, y, z);
    return;
  }

  /* Do all cases not handled above:
     (1) a == other, b == 0.0 - user should have called N_VScale
     (2) a == 0.0, b == other - user should have called N_VScale
     (3) a,b == other, a !=b, a != -b */
  
  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++) 
    *zd++ = a * (*xd++) + b * (*yd++);
}


void N_VConst(real c, N_Vector z)
{
  integer i, N;
  real *zd;

  N = z->length;
  zd = z->data;

  for (i=0; i < N; i++) 
    *zd++ = c;
}


void N_VProd(N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;

  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = (*xd++) * (*yd++);
}


void N_VDiv(N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;

  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = (*xd++) / (*yd++);
}


void N_VScale(real c, N_Vector x, N_Vector z)
{
  integer i, N;
  real *xd, *zd;

  if (z == x) {       /* BLAS usage: scale x <- cx */
    VScaleBy(c, x);
    return;
  }

  if (c == ONE) {
    VCopy(x, z);
  } else if (c == -ONE) {
    VNeg(x, z);
  } else {
    N = x->length;
    xd = x->data;
    zd = z->data;
    for (i=0; i < N; i++) *zd++ = c * (*xd++);
  }
}


void N_VAbs(N_Vector x, N_Vector z)
{
  integer i, N;
  real *xd, *zd;

  N = x->length;
  xd = x->data;
  zd = z->data;

  for (i=0; i < N; i++, xd++, zd++)
    *zd = ABS(*xd);
}


void N_VInv(N_Vector x, N_Vector z)
{
  integer i, N;
  real *xd, *zd;

  N = x->length;
  xd = x->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = ONE / (*xd++);
}


void N_VAddConst(N_Vector x, real b, N_Vector z)
{
  integer i, N;
  real *xd, *zd;
  
  N = x->length;
  xd = x->data;
  zd = z->data;
  
  for (i=0; i < N; i++) *zd++ = (*xd++) + b; 
}


real N_VDotProd(N_Vector x, N_Vector y)
{
  integer i, loclen;
  real sum = ZERO, *xd, *yd, gsum;
  machEnvType machenv;

  loclen = x->length;
  xd = x->data;
  yd = y->data;
  machenv = x->machEnv;

  for (i=0; i < loclen; i++) sum += xd[i] * yd[i];

  gsum = PVecAllReduce(sum, 1, machenv);
  return(gsum);
}


real N_VMaxNorm(N_Vector x)
{
  integer i, N;
  real max = ZERO, *xd, gmax;
  machEnvType machenv;

  N = x->length;
  xd = x->data;
  machenv = x->machEnv;

  for (i=0; i < N; i++, xd++) {
    if (ABS(*xd) > max) max = ABS(*xd);
  }
   
  gmax = PVecAllReduce(max, 2, machenv);
  return(gmax);
}


real N_VWrmsNorm(N_Vector x, N_Vector w)
{
  integer i, N, N_global;
  real sum = ZERO, prodi, *xd, *wd, gsum;
  machEnvType machenv;

  N = x->length;
  N_global = x->global_length;
  xd = x->data;
  wd = w->data;
  machenv = x->machEnv;

  for (i=0; i < N; i++) {
    prodi = (*xd++) * (*wd++);
    sum += prodi * prodi;
  }

  gsum = PVecAllReduce(sum, 1, machenv);
  return(RSqrt(gsum / N_global));
}

real N_VMin(N_Vector x)
{
  integer i, N;
  real min, *xd, gmin;
  machEnvType machenv;

  N = x->length;
  xd = x->data;
  min = xd[0];
  machenv = x->machEnv;

  for (i=0; i < N; i++, xd++) {
    if ((*xd) < min) min = *xd;
  }

  gmin = PVecAllReduce(min, 3, machenv);
  return(gmin);
}


real N_VWL2Norm(N_Vector x, N_Vector w)
{
  integer i, N, N_global;
  real sum = ZERO, prodi, *xd, *wd, gsum;
  machEnvType machenv;

  N = x->length;
  N_global = x->global_length;
  xd = x->data;
  wd = w->data;
  machenv = x->machEnv;

  for (i=0; i < N; i++) {
    prodi = (*xd++) * (*wd++);
    sum += prodi * prodi;
  }

  gsum = PVecAllReduce(sum, 1, machenv);
  return(RSqrt(gsum));
}


real N_VL1Norm(N_Vector x)
{
  integer i, N;
  real sum = ZERO, gsum, *xd;
  machEnvType machenv;


  N = x->length;
  xd = x->data;
  machenv = x->machEnv;

  for (i=0; i<N; i++,xd++) sum += ABS(*xd);

  gsum = PVecAllReduce(sum, 1, machenv);

  return(gsum);
}


void N_VCompare(real c, N_Vector x, N_Vector z)
{
  integer i, N;
  real *xd, *zd;
  
  N = x->length;
  xd = x->data;
  zd = z->data;
  
  for (i=0; i < N; i++, xd++, zd++) {
    *zd = (ABS(*xd) >= c) ? ONE : ZERO;
  }
}


boole N_VInvTest(N_Vector x, N_Vector z)
{
  integer i, N;
  real *xd, *zd, val, gval;
  machEnvType machenv;

  N = x->length;
  xd = x->data;
  zd = z->data;
  machenv = x->machEnv;

  val = 1;
  for (i=0; i < N; i++) {
    if (*xd == ZERO) 
      val = 0;
    else
      *zd++ = ONE / (*xd++);
  }

  gval = PVecAllReduce(val, 3, machenv);
  if (gval == 0)
    return(FALSE);
  else
    return(TRUE);
}


boole N_VConstrProdPos(N_Vector c, N_Vector x)
{
  integer i, N;
  boole test;
  real  *xd, *cd;
  machEnvType machenv;

  N =  x->length;
  xd = x->data;
  cd = c->data;
  machenv = x->machEnv;

  test = TRUE;
  for (i=0; i < N; i++, xd++,cd++) {
    if (*cd !=ZERO){
      if(*xd * *cd <= ZERO) {
	test = FALSE;
        break;
      }
    }
  }
  return((boole)PVecAllReduce((int)test,3,machenv));
}

 
void N_VPrint(N_Vector x)
{
  integer i, N;
  real *xd;

  N = x->length;
  xd = x->data;

  for (i=0; i < N; i++) printf("%g\n", *xd++);

  printf("\n");
}


/***************** Private Helper Functions **********************/

static real PVecAllReduce(real d, int op, machEnvType machEnv)
{
  /* This function does a global reduction.  The operation is
       sum if op = 1,
       max if op = 2,
       min if op = 3.
     The operation is over all processors in the group defined by
     the parameters within machEnv. */

  MPI_Comm comm;
  real out;

  comm = machEnv->comm;

  switch (op) {
   case 1: MPI_Allreduce(&d, &out, 1, PVEC_REAL_MPI_TYPE, MPI_SUM, comm);
           break;

   case 2: MPI_Allreduce(&d, &out, 1, PVEC_REAL_MPI_TYPE, MPI_MAX, comm);
           break;

   case 3: MPI_Allreduce(&d, &out, 1, PVEC_REAL_MPI_TYPE, MPI_MIN, comm);
           break;

   default: break;
  }

  return(out);
}


static void VCopy(N_Vector x, N_Vector z)
{
  integer i, N;
  real *xd, *zd;

  N = x->length;
  xd = x->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = *xd++; 
}


static void VSum(N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;

  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = (*xd++) + (*yd++);
}


static void VDiff(N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;
 
  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = (*xd++) - (*yd++);
}


static void VNeg(N_Vector x, N_Vector z)
{
  integer i, N;
  real *xd, *zd;

  N = x->length;
  xd = x->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = -(*xd++);
}


static void VScaleSum(real c, N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;

  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = c * ((*xd++) + (*yd++));
}


static void VScaleDiff(real c, N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;

  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = c * ((*xd++) - (*yd++));
}


static void VLin1(real a, N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;

  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = a * (*xd++) + (*yd++);
}


static void VLin2(real a, N_Vector x, N_Vector y, N_Vector z)
{
  integer i, N;
  real *xd, *yd, *zd;

  N = x->length;
  xd = x->data;
  yd = y->data;
  zd = z->data;

  for (i=0; i < N; i++)
    *zd++ = a * (*xd++) - (*yd++);
}

static void Vaxpy(real a, N_Vector x, N_Vector y)
{
  integer i, N;
  real *xd, *yd;

  N = x->length;
  xd = x->data;
  yd = y->data;

  if (a == ONE) {
    for (i=0; i < N; i++)
      *yd++ += (*xd++);
    return;
  }

  if (a == -ONE) {
    for (i=0; i < N; i++)
      *yd++ -= (*xd++);
    return;
  }    

  for (i=0; i < N; i++)
    *yd++ += a * (*xd++);
}

static void VScaleBy(real a, N_Vector x)
{
  integer i, N;
  real *xd;

  N = x->length;
  xd = x->data;

  for (i=0; i < N; i++)
    *xd++ *= a;
}
