/******************************************************************
 *                                                                *
 * File          : sensitivity.c                                  *
 * Last Modified : 20 August 1999                                 *
 *----------------------------------------------------------------*
 * This is the implementation file for computing the sensitivity  *
 * of the ODE solution y(t, p) with respect to the parameters     *
 * in p.                                                          *
 *                                                                *
 ******************************************************************/


/************************************************************/
/******************* BEGIN Imports **************************/
/************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "llnltyps.h"
#include "nvector.h"
#include "sensitivity.h"

/************************************************************/
/******************** END Imports ***************************/
/************************************************************/

/************************************************************/
/************** BEGIN Private Constants *********************/
/************************************************************/

#define ZERO      RCONST(0.0)  /* real 0.0   */
#define HALF      RCONST(0.5)  /* real 0.5   */
#define ONE       RCONST(1.0)  /* real 1.0   */

/***************************************************************/
/***************** BEGIN Error Messages ************************/
/***************************************************************/

/* SensCVodeMalloc Error Messages */

#define SensCVM         "SensCVodeMalloc-- "

#define MSG_BAD_NY       SensCVM "ny=%ld < 1 illegal.\n\n"
#define MSG_BAD_NS       SensCVM "ns=%ld < 0 illegal.\n\n"
#define MSG_P_NULL       SensCVM "p=NULL illegal.\n\n"
#define MSG_PBAR_NULL    SensCVM "pbar=NULL illegal.\n\n"

/***************************************************************/
/****************** END Error Messages *************************/
/***************************************************************/

/************************************************************/
/************** END Private Constants ***********************/
/************************************************************/


/**************************************************************/
/********* BEGIN Private Helper Functions Prototypes **********/
/**************************************************************/

void FDQ(integer Ntotal, real t, N_Vector u, N_Vector udot, 
                void *s_data);

/**************************************************************/
/********** END Private Helper Functions Prototypes ***********/
/**************************************************************/

/* AD Helper Function Prototypes - added by Paul */

void ad_Malloc(SensData sens, int n, machEnvType m);
void ad_Free(SensData sens);

/* end AD Helper Function Prototypes */

/******************** SensCVodeMalloc ***************************

 SensCVodeMalloc initializes a block of sensitivity data.
 It also allocates and initializes memory for the ODE problem. 
 All problem specification inputs are checked for errors. If any
 error occurs during initialization, it is reported to the file
 whose file pointer is errfp and NULL is returned. Otherwise, the
 pointer to successfully initialized problem memory is returned.
 
*****************************************************************/

void *SensCVodeMalloc(integer Ntotal, RhsFn f, real t0, N_Vector y0, 
                      int lmm, int iter, int itol, real *reltol, 
                      void *abstol, void *f_data, FILE *errfp,
                      boole optIn, long int iopt[], real ropt[], 
                      void *machEnv, int ny, int ns, void *p, 
                      void *pbar, real rhomax)
{
  SensData sdata;
  machEnvType machenv;
  void *cv_mem;
  FILE *fp;

  /* Check for legal input parameters */

  fp = (errfp == NULL) ? stdout : errfp;

  if (ny <= 0) {
    fprintf(fp, MSG_BAD_NY, ny);
    return(NULL);
  }
  if (ns < 0) {
    fprintf(fp, MSG_BAD_NS, ns);
    return(NULL);
  }
  if (p == NULL) {
    fprintf(fp, MSG_P_NULL);
    return(NULL);
  }
  if (pbar == NULL) {
    fprintf(fp, MSG_PBAR_NULL);
    return(NULL);
  }

  sdata = (SensData) malloc(sizeof(*sdata));
  machenv = (machEnvType) machEnv;
  sdata->ny = ny;
  sdata->ns = ns;
  sdata->f_data = f_data;
  sdata->p = p;
  sdata->pbar = pbar;
  sdata->rhomax = rhomax;
  sdata->f_ptr = f; 
  sdata->ftemp = N_VNew(ny, machenv);
  sdata->ytemp = N_VNew(ny, machenv);
  cv_mem = CVodeMalloc(Ntotal, FDQ, t0, y0, lmm, iter, itol, reltol,
                       abstol, sdata, errfp, optIn, iopt, ropt, machEnv);
  sdata->cv_mem = cv_mem;

  /* next line added by Paul */
  ad_Malloc(sdata,ny,machenv);

  return(cv_mem);
}


/********************* SensCVodeFree ******************************

 This routine frees the block of sensitivity data and problem memory 
 allocated by SensCVodeMalloc. Such memory includes all the vectors 
 allocated by CVAllocVectors, and the memory lmem for the linear solver 
 (deallocated by a call to lfree).

*******************************************************************/

void SensCVodeFree(void *cvode_mem)
{
  CVodeMem cv_mem;
  SensData sdata;

  cv_mem = (CVodeMem) cvode_mem;
  sdata = (SensData) cv_mem->cv_f_data;

  /* next line added by Paul */
  ad_Free(sdata);

  N_VFree(sdata->ftemp);
  N_VFree(sdata->ytemp);
  free(sdata);

  CVodeFree(cvode_mem);
}


/********************* SensSetAtol ********************************

 This routine sets the absolute error tolerances for the sensitivity
 vectors. The absolute error tolerances for scaled sensitivity vector 
 i is the same as the absolute error tolerances for the state vector.

*******************************************************************/

void SensSetAtol(N_Vector atol, integer ns)
{
  N_Vector *atolsub;
  int i;

  atolsub = (N_Vector *) atol->subptr;

  /* Load state error tolerances as sensitivity error tolerances */
  for (i = 1; i <= ns; i++)
    N_VScale(ONE, atolsub[0], atolsub[i]);
}

/***************************************************************/
/********** END Exported Functions Implementation **************/
/***************************************************************/
